#!/usr/bin/env python
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>Custom Font Tools>FaceMaker'
#
# The first option creates new letter faces from over-sized visible layer/s
# The second option clears the contents of the old character layer before copying and pasting the new letter face
# The third and fourth options set the mode (legacy and new) to be used when copying and pasting to generate the new font face
# The fifth option removes non-font character layers from the image - apart from any layer name containing the word 'legend'
# The sixth option will re-position all layers to the origin of their parent image (Option 4 does this too)
# The seventh option will re-order the layers in the custom font file according to the following order
#		space.png (or  .ping)
#		Capital letters (A-Z)
#		Lower case letters (a-z)
#		Numbers (0-9)
#		Basic maths characters
# 		Punctuation characters
#		Other keyboard characters
#		Special characters (Pound Sterling sign, Copyright Sign, Registered Trademark sign)
# You can change this order by altering the appropriate lines (currently beginning at line no 130) of the plug-in
# The eighth option will display a message box listing all the characters (ie layers) defined in the custom font
# The ninth option defines the string to be used to append to single letter layer and/or path names
# The tenth option will add .png to all the layers with one character names ie  A to A.png
# The eleventh option will add .png to all the paths with one character names
# The twelfth option will combine the characters from the xcf custom font file selected, with the file selection tool below, with the current characters
# The thirteenth option is the xcf file selection tool
# The fourteenth option trims pixels from the character

# Version 0.01 Fixed adding .png to single letter layers and paths with special characters ie unicode characters
# Version 1.02 Adds an option to create a new face for a custom font from additional/decorative visible layers
# Version 1.03 Re-position space layer, if one exists, along with other letter layers and adds the option to remove non-font-character layers from the image
# Version 1.04 Trim and feather pixels from letter faces to remove halo effect
# Version 1.05 Fixed bug where character layer is not at image origin (and so doesn't copy new face correctly): Adds option to use different modes when copying from visible layers to font face
# Version 1.06 Added option to re-position all layers to the origin of their parent image ie with x and y co-ordinates set to zero: Re-organized dialog and filter structure so that all options can be set without any issues
# Version 1.07 Added option to clear the existing letter face before copying: Re-ordered flow of filter (yet again); Added options to combine two custom font files
# Version 1.08 Fixed potential bug when layer is bigger than image
#              Changed filter location
# Version 1.09 Added mode choice when combining two custom fonts
# Version 1.10 Added fixed position option for creating new characters from a visible layer
# Version 1.11 Adds an option to trim pixels from each character; changed plug-in name to Custom Font Layer Utilities
# Version 1.12 Adds an option to add lower-case characters to the custom font by copying and scaling, as a percentage, the upper-case characters (when the custom font has a path called base_line.png)
# Version 1.13 Adds an option to sandwich a character between a top layer and a bottom layer - optionally setting the layer size to the full width of the font image size
# Version 0.01 Custom Font Face Maker 
#              Removed options not used in making character faces from visible layer
#              Added option for using all non-character layers to make new character faces from sequentially
#              Added option to use non-character layers randomly
#              Added option to use non-character layers with groups according to non-character layer name (upper/lower/number/punctuation/other)
# Version 0.02 adds checking for unicode symbols, reworked using character group faces
# Version 0.03 general improvements
# Version 0.04 Re-worked use of random character faces
# Version 0.05 Added use of smaller increments for growing the letter size
# Version 0.06 Re-worked trimming halo from character outlines
# Version 0.07 Fixed mode issue
# Version 0.08 general improvements
# Version 0.09 Added opacity slider
# Version 0.10 Added preview to new image option, character group option and fixed a bug that had been introduced when copying characters from another file with fixed position set

from gimpfu import *
from array import array
from gtk import gdk
import math
import random

w = gdk.screen_width()
h = gdk.screen_height()

# Some global colours we may need to use

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)
darkgrey = gimpcolor.RGB(50, 50, 50)
midgrey = gimpcolor.RGB(125, 125, 125)
lightgrey = gimpcolor.RGB(170, 170, 170)
verylightgrey = gimpcolor.RGB(222, 222, 222)
red = gimpcolor.RGB(220,10,10)	


def do_stuff(image, layer, do_new_letters, fixed_position, grow, many_faces, many_rnd, grouping, clear_contents, old_mode, new_mode, opacity, 
            do_merge_files, xcffilename, do_preview):

    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)	
	
    pdb.gimp_context_set_feather(False)
    #pdb.gimp_context_set_antialias(False)
	
    do_sandwich = False
    add_case = False
    do_layer_names = False
    do_path_names = False
    do_sort = False
    do_place_at_origin = False
    do_remove_non_letter_layers = False
    do_layer_info = False
    trim = 0

	
    if do_new_letters:
	
	
        if do_preview:
		
            preview_other, preview_case, preview_shape, preview_vowel = 0, 0, 0, 0
		
            preview_im_width, preview_im_height = image.width, image.height
            preview_image = pdb.gimp_image_new(preview_im_width, preview_im_height, 0)
            temp_preview_layer = pdb.gimp_layer_new(preview_image, preview_im_width, preview_im_height, 1, "temp preview layer", 100, 28)
            pdb.gimp_image_insert_layer(preview_image, temp_preview_layer, None, -1)
            pdb.gimp_image_remove_layer(preview_image, temp_preview_layer)
            preview_display = pdb.gimp_display_new(preview_image)
            # Set start undo point for preview image
            pdb.gimp_image_undo_group_start(preview_image)	
            
		
        # define some character groupings
        upper_case, lower_case, number_case, punctuation_case = "ABCDEFGHIJKLMNOPQRSTUVWXYZ", "abcdefghijklmnopqrstuvwxyz", "0123456789", ".,:;'!?\""
        vowel_case, consonant_case = "AEIOUaeiou","BCDFGHJKLMNPQRSTVWXYZbcdfghjklmnpqrstvwxyz"
        straight_case, round_case, mixed_case = "AEFHIKLMNTVWXYZikltvwxz", "CGOQSaceos", "BDJPRUbdfghjmnpqruy"
		
		
        # set groups
        if grouping == 0:
            group = "none"
        elif grouping == 1:
            group = "case"
        elif grouping == 2:
            group = "shape"
        elif grouping == 3:
            group = "vowel"
		
        # define some preview character values
        preview_layer_pos = -1
        preview_upper_pos = -1
        preview_lower_pos = -1
        preview_number_pos = -1
        preview_punctuation_pos = -1
        preview_vowel_pos = -1
        preview_consonant_pos = -1
        preview_straight_pos = -1
        preview_round_pos = -1
        preview_mixed_pos = -1
		
        mode = fn_get_mode(old_mode, new_mode)
        im_width = image.width
        im_height = image.height
    
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
				
        visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
        multi_layer_list = "No non-character layers."
		
        num_multi_layers = 0
        while visi_loop < num_layers:# loop through layers and store any non-character layers for creating new faces
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if visi_name == "space.png" or visi_name == " .png": has_a_space = True
            if fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name): # check for custom font layer now allows unicode symbols
                if first_letter_layer < 0: first_letter_layer = visi_loop
            else:
                pdb.gimp_image_select_item(image, 2, layer_names_list[visi_loop])
                is_empty = pdb.gimp_selection_is_empty(image)
                if is_empty == False:
                    if other_layers == 0:
                        
                        multi_layer_list = [visi_name]
                    else:
                        
                        multi_layer_list.append(visi_name)
                    other_layers += 1
                    
                    		
            visi_loop += 1

        if str(multi_layer_list) == "No non-character layers.": 
            num_multi_layers = 0
        else:
            num_multi_layers = len(multi_layer_list)
        if many_faces:
            if num_multi_layers < 2: many_faces = False
			
        if many_faces: copy_multi_layer_list = fn_copy_list(multi_layer_list)
			
        visi_loop = 0
        real_name_layer_list = ["dummy"] # list to hold actual character names 
        while visi_loop < num_layers and other_layers > 0: # turn off visibility of letter layers
            real_visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = real_visi_name.decode('utf-8')
            if visi_name == "space.png" or visi_name == " .png": 
                has_a_space = True
                space_loop = visi_loop
            if len(visi_name) == 1 or len(visi_name) == 5 and ".png" in visi_name:
                pdb.gimp_item_set_visible(layer_names_list[visi_loop],False) 
                if real_visi_name<>"space.png" and real_visi_name<>" .png": # add some preview image variables
                    real_name_layer_list.append(real_visi_name)
                    if preview_layer_pos == -1: preview_layer_pos = visi_loop
                    if real_visi_name[0] in upper_case:
                        if preview_upper_pos == -1: preview_upper_pos = visi_loop
                    if real_visi_name[0] in lower_case:
                        if preview_lower_pos == -1: preview_lower_pos = visi_loop
                    if real_visi_name[0] in number_case:
                        if preview_number_pos == -1: preview_number_pos = visi_loop
                    if real_visi_name[0] in punctuation_case:
                        if preview_punctuation_pos == -1: preview_punctuation_pos = visi_loop
                    if real_visi_name[0] in vowel_case:
                        if preview_vowel_pos == -1: preview_vowel_pos = visi_loop
                    if real_visi_name[0] in consonant_case:
                        if preview_consonant_pos == -1: preview_consonant_pos = visi_loop
                    if real_visi_name[0] in straight_case:
                        if preview_straight_pos == -1: preview_straight_pos = visi_loop
                    if real_visi_name[0] in round_case:
                        if preview_round_pos == -1: preview_round_pos = visi_loop
                    if real_visi_name[0] in mixed_case:
                        if preview_mixed_pos == -1: preview_mixed_pos = visi_loop
					
            visi_loop += 1
        #real_name_layer_list.remove("dummy")
        #pdb.gimp_message(str(real_name_layer_list) + " preview_punctuation_pos = " + str(preview_punctuation_pos) + " layer name = " + str(real_name_layer_list[preview_punctuation_pos]) )
			
        if other_layers > 0: # start copying decor onto characters
            
            if do_preview:
                ### check for layers to copy to preview image/insert layers in preview layer/turn off layer visibility
                if group == "none":
                    preview_layer = layer_names_list[preview_layer_pos]
                    preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                    temp_preview_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                    pdb.gimp_image_insert_layer(preview_image, temp_preview_layer, None, -1)
                    success = pdb.gimp_edit_copy(preview_layer)
                    floating_sel = pdb.gimp_edit_paste(temp_preview_layer, True)
                    pdb.gimp_floating_sel_anchor(floating_sel)
                    pdb.gimp_item_set_visible(temp_preview_layer,False) 
                else:
                    if group == "case":
                        if preview_punctuation_pos > -1:
                            preview_layer = layer_names_list[preview_punctuation_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_punctuation_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_punctuation_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_punctuation_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_punctuation_layer,False) 
                        if preview_number_pos > -1:
                            preview_layer = layer_names_list[preview_number_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_number_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_number_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_number_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_number_layer,False) 
                        if preview_lower_pos > -1:
                            preview_layer = layer_names_list[preview_lower_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_lower_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_lower_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_lower_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_lower_layer,False) 
                        if preview_upper_pos > -1:
                            preview_layer = layer_names_list[preview_upper_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_upper_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_upper_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_upper_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_upper_layer,False)
						
                    if group == "vowel":
                        if preview_consonant_pos > -1:
                            preview_layer = layer_names_list[preview_consonant_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_consonant_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_consonant_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_consonant_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_consonant_layer,False)
                        if preview_vowel_pos > -1:
                            preview_layer = layer_names_list[preview_vowel_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_vowel_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_vowel_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_vowel_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_vowel_layer,False)
							
                    if group == "shape":
                        if preview_mixed_pos > -1:
                            preview_layer = layer_names_list[preview_mixed_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_mixed_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_mixed_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_mixed_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_mixed_layer,False)
                        if preview_round_pos > -1:
                            preview_layer = layer_names_list[preview_round_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_round_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_round_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_round_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_round_layer,False)
                        if preview_straight_pos > -1:
                            preview_layer = layer_names_list[preview_straight_pos]
                            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
                            #pdb.gimp_message(" preview_layer_name = " + str(preview_layer_name) )
                            temp_straight_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
                            pdb.gimp_image_insert_layer(preview_image, temp_straight_layer, None, -1)
                            success = pdb.gimp_edit_copy(preview_layer)
                            floating_sel = pdb.gimp_edit_paste(temp_straight_layer, True)
                            pdb.gimp_floating_sel_anchor(floating_sel)
                            pdb.gimp_item_set_visible(temp_straight_layer,False)
							
                pdb.gimp_image_resize_to_layers(preview_image)
            
            if do_preview:
                ####
                num_preview_layers, preview_layer_ids = pdb.gimp_image_get_layers(preview_image)
                preview_layer_names_list = preview_image.layers
                preview_loop = 0
				
                while preview_loop < num_preview_layers:
                    preview_name = pdb.gimp_item_get_name(preview_layer_names_list[preview_loop])
                    preview_name = preview_name.decode('utf-8')
					
					
                    layer_width = preview_layer_names_list[preview_loop].width
                    if layer_width > preview_image.width: layer_width = preview_image.width
                    layer_height = preview_layer_names_list[preview_loop].height
                    if layer_height > preview_image.height: layer_height = preview_image.height
					
					
                    if fixed_position: # sort out x y positions to copy from
                        random_x = 0
                        if layer_width < preview_image.width: random_x = int((preview_image.width - layer_width)/2)
                        random_y = 0
                        if layer_height < preview_image.height: random_y = int((preview_image.height - layer_height)/2)
                    else:
                        random.seed()
                        random_x = random.randrange(0, (preview_image.width + 1) - layer_width)
                        random.seed()
                        random_y = random.randrange(0, (preview_image.height + 1) - layer_height)
						
						
                    # make a duplicate of the current character layer and insert
                    layer_position = pdb.gimp_image_get_item_position(preview_image, preview_layer_names_list[preview_loop])
                    layer_copy = pdb.gimp_layer_copy(preview_layer_names_list[preview_loop], True)
                    pdb.gimp_image_insert_layer(preview_image, layer_copy, None, layer_position)
                    # set random position to copy visible layers 
                    #layer_copy.set_offsets(random_x, random_y)
					
                    # select the contents of the layer and clear contents
                    pdb.gimp_image_select_item(preview_image, 2, layer_copy)
                    if grow>0:
                        pdb.gimp_selection_grow(preview_image, grow)
                    pdb.gimp_drawable_edit_clear(layer_copy)
					
                    # cancel use of multi decorative layers if character groups are labelled 
                    if group<>"none":
                    #if "other" in str(multi_layer_list) or "upper" in str(multi_layer_list) or "lower" in str(multi_layer_list) or "number" in str(multi_layer_list) or "punctuation" in str(multi_layer_list): 
                        many_faces = False
						
						
                    if many_faces:
                        if many_rnd: # randomize choice of faces
                            random.seed()
                            random_x = random.randrange(0, (preview_image.width + 1) - layer_width)
							
                            random_layer = random.choice(copy_multi_layer_list)
                            multi_target = multi_layer_list.index(random_layer)
							
							
                            random.seed()
					
				
                            if len(copy_multi_layer_list)>0:
                                copy_multi_layer_list.remove(random_layer)
					
                            if len(copy_multi_layer_list)<1:
                                copy_multi_layer_list = fn_copy_list(multi_layer_list)
                                random.shuffle(copy_multi_layer_list)

                        else:
                            multi_target = visi_loop % num_multi_layers
						
                        multi_loop = 0
                        while multi_loop < num_multi_layers: # set visibility of current copy layer in original image 
                            visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                            if multi_loop == multi_target:
                                pdb.gimp_item_set_visible(visible_layer, True)
                            else:
                                pdb.gimp_item_set_visible(visible_layer, False)
                            multi_loop += 1
							
                    else:
                        multi_target = 0 # topmost layer used as face for any layer not catered for by the character groups below
                        multi_loop = 0
                        while multi_loop < num_multi_layers:
                            visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                            if multi_loop == multi_target:
                                pdb.gimp_item_set_visible(visible_layer, True)
                            else:
                                pdb.gimp_item_set_visible(visible_layer, False)
                            multi_loop += 1
							
                        if "other" in str(multi_layer_list) and preview_name[0] not in punctuation_case and preview_name[0] not in number_case and preview_name[0] not in lower_case and preview_name[0] not in upper_case and preview_name[0] not in vowel_case and preview_name[0] not in consonant_case and preview_name[0] not in straight_case and preview_name[0] not in round_case and preview_name[0] not in mixed_case: 
                            counter = 0
                            while counter < num_multi_layers:
                                if "other" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
									
                                multi_loop += 1
					
                        elif preview_name[0] in upper_case and "upper" in str(multi_layer_list) and group == "case": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "upper" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
								
                        elif preview_name[0] in lower_case and "lower" in str(multi_layer_list) and group == "case": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "lower" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
							
					
                        elif preview_name[0] in number_case and "number" in str(multi_layer_list) and group == "case":
                            counter = 0
                            while counter < num_multi_layers:
                                if "number" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
                        elif preview_name[0] in punctuation_case and "punctuation" in str(multi_layer_list) and group == "case": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "punctuation" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
							
                        elif preview_name[0] in vowel_case and "vowel" in str(multi_layer_list) and group == "vowel": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "vowel" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
							
                        elif preview_name[0] in consonant_case and "consonant" in str(multi_layer_list) and group == "vowel": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "consonant" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
																
														
                        elif preview_name[0] in straight_case and "straight" in str(multi_layer_list) and group == "shape": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "straight" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
								
                        elif preview_name[0] in round_case and "round" in str(multi_layer_list) and group == "shape": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "round" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
								
							
                        elif preview_name[0] in mixed_case and "mixed" in str(multi_layer_list) and group == "shape": 
                            counter = 0
                            while counter < num_multi_layers:
                                if "mixed" in str(multi_layer_list[counter]):
                                    multi_target = counter
                                counter += 1
								
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
					
                    # copy the contents of the visible layers into the selection
                    pdb.gimp_selection_grow(preview_image, 1)
                    success = pdb.gimp_edit_copy_visible(image)
					
                    # turn on layer visibility
                    pdb.gimp_item_set_visible(layer_copy, True) 
					
                    # paste the selection into the floating layer
                    floating_sel = pdb.gimp_edit_paste(layer_copy,True)
                    # and anchor it
                    pdb.gimp_floating_sel_anchor(floating_sel)
                    # set the layer mode as per the dialog UI when the contents of the original character layer have not been cleared
                    if not clear_contents: pdb.gimp_layer_set_mode(layer_copy, mode)
                    # Re-position the copied layer
                    layer_copy.set_offsets(0, 0)
					
					
                    # use the original layer to copy the alpha selection before clearing the original character
                    if clear_contents: 
					
                        pdb.gimp_image_select_item(preview_image, 2, layer_copy)
                        if grow>0:
                            pdb.gimp_selection_grow(preview_image, grow)
                        else:
                            proc_trim_away(preview_image, layer_copy, preview_layer_names_list[preview_loop], 1)
					
					
                    # Make the original character layer visible before we merge down
					
                    pdb.gimp_item_set_visible( preview_layer_names_list[preview_loop], True)
                    # option to clear the original character
                    if clear_contents: 
                        pdb.gimp_selection_none(preview_image)
                        pdb.gimp_drawable_edit_clear(preview_layer_names_list[preview_loop])

						
						
                    if not clear_contents: pdb.gimp_layer_set_opacity(layer_copy, opacity)
                    # Merge down the copied layer to the character layer
                    pdb.gimp_image_merge_down(preview_image, layer_copy, 0)
                    # Re-set the id of the layer
                    preview_layer_names_list[preview_loop] = pdb.gimp_image_get_active_layer(preview_image)
                    # Remove annoying halo
                    #if not clear_contents: proc_trim_away(image, layer_names_list[visi_loop], layer_names_list[visi_loop], 1)
                    # Turn off layer visibility
                    pdb.gimp_item_set_visible( preview_layer_names_list[preview_loop], False) 
				
						
                    '''		
                    '''
							
					
					
                    preview_loop += 1
                    pdb.gimp_selection_none(preview_image)
				
                # Deselect
                pdb.gimp_selection_none(preview_image)
				
                # Turn on preview image first layer visibility
                pdb.gimp_item_set_visible( preview_layer_names_list[0], True) 
                pdb.gimp_image_set_active_layer(preview_image, preview_layer_names_list[0])
                # Set end undo point for preview image
                pdb.gimp_image_undo_group_end(preview_image)
				
				
            else:
                # apply to all characters in original image
                visi_loop = 0
                while visi_loop < num_layers:
                    visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
                    visi_name = visi_name.decode('utf-8')
                    #if not fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name): pdb.gimp_message(visi_name)
                    if fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name): # cater for unicode symbol characters
                        # make sure that existing layers are positioned at the origin of the image
                        layer_names_list[visi_loop].set_offsets(0, 0)
                        # prepare to move layer to random position within the confines of the image size
                        layer_width = layer_names_list[visi_loop].width
                        if layer_width > im_width: layer_width = im_width
                        layer_height = layer_names_list[visi_loop].height
                        if layer_height > im_height: layer_height = im_height

                        if fixed_position: # sort out x y positions to copy from
                            random_x = 0
                            if layer_width < im_width: random_x = int((im_width - layer_width)/2)
                            random_y = 0
                            if layer_height < im_height: random_y = int((im_height - layer_height)/2)
                        else:
                            random.seed()
                            random_x = random.randrange(0, (im_width + 1) - layer_width)
                            random.seed()
                            random_y = random.randrange(0, (im_height + 1) - layer_height)
					
                        # make a duplicate of the current character layer and insert
                        layer_position = pdb.gimp_image_get_item_position(image, layer_names_list[visi_loop])
                        layer_copy = pdb.gimp_layer_copy(layer_names_list[visi_loop], True)
                        pdb.gimp_image_insert_layer(image, layer_copy, None, layer_position)
                        # set random position to copy visible layers 
                        layer_copy.set_offsets(random_x, random_y)
                    
				
                        # select the contents of the layer and clear contents
                        pdb.gimp_image_select_item(image, 2, layer_copy)
                        if grow>0:
                            pdb.gimp_selection_grow(image, grow)
                        pdb.gimp_drawable_edit_clear(layer_copy)
					
                        # cancel use of multi decorative layers if character groups are labelled 
                        if group<>"none":
                        #if "other" in str(multi_layer_list) or "upper" in str(multi_layer_list) or "lower" in str(multi_layer_list) or "number" in str(multi_layer_list) or "punctuation" in str(multi_layer_list): 
                            many_faces = False
                        
						
                        if many_faces:
                            if many_rnd: # randomize choice of faces
                                random.seed()
                                random_x = random.randrange(0, (im_width + 1) - layer_width)
							
                                random_layer = random.choice(copy_multi_layer_list)
                                multi_target = multi_layer_list.index(random_layer)
							
                                #multi_target = (random.randrange(0, num_multi_layers) )
							
                                random.seed()
					
				
                                if len(copy_multi_layer_list)>0:
                                    copy_multi_layer_list.remove(random_layer)
					
                                if len(copy_multi_layer_list)<1:
                                    copy_multi_layer_list = fn_copy_list(multi_layer_list)
                                    random.shuffle(copy_multi_layer_list)

                            else:
                                multi_target = visi_loop % num_multi_layers
							
                            multi_loop = 0
                            while multi_loop < num_multi_layers: # set visibility of current copy layer
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
							
                        else:
                            multi_target = 0 # topmost layer used as face for any layer not catered for by the character groups below
                            multi_loop = 0
                            while multi_loop < num_multi_layers:
                                visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                if multi_loop == multi_target:
                                    pdb.gimp_item_set_visible(visible_layer, True)
                                else:
                                    pdb.gimp_item_set_visible(visible_layer, False)
                                multi_loop += 1
							
                            if "other" in str(multi_layer_list) and visi_name[0] not in punctuation_case and visi_name[0] not in number_case and visi_name[0] not in lower_case and visi_name[0] not in upper_case and visi_name[0] not in vowel_case and visi_name[0] not in consonant_case and visi_name[0] not in straight_case and visi_name[0] not in round_case and visi_name[0] not in mixed_case: 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "other" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
						
                            elif visi_name[0] in upper_case and "upper" in str(multi_layer_list) and group == "case": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "upper" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
							
                            elif visi_name[0] in lower_case and "lower" in str(multi_layer_list) and group == "case": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "lower" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
							
                            elif visi_name[0] in number_case and "number" in str(multi_layer_list) and group == "case":
                                counter = 0
                                while counter < num_multi_layers:
                                    if "number" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
							
                            elif visi_name[0] in punctuation_case and "punctuation" in str(multi_layer_list) and group == "case": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "punctuation" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
								
							
                            elif visi_name[0] in vowel_case and "vowel" in str(multi_layer_list) and group == "vowel": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "vowel" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
								
							
                            elif visi_name[0] in consonant_case and "consonant" in str(multi_layer_list) and group == "vowel": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "consonant" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
								
														
                            elif visi_name[0] in straight_case and "straight" in str(multi_layer_list) and group == "shape": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "straight" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
								
							
                            elif visi_name[0] in round_case and "round" in str(multi_layer_list) and group == "shape": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "round" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
								
							
                            elif visi_name[0] in mixed_case and "mixed" in str(multi_layer_list) and group == "shape": 
                                counter = 0
                                while counter < num_multi_layers:
                                    if "mixed" in str(multi_layer_list[counter]):
                                        multi_target = counter
                                    counter += 1
								
                                multi_loop = 0
                                while multi_loop < num_multi_layers:
                                    visible_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[multi_loop])
                                    if multi_loop == multi_target:
                                        pdb.gimp_item_set_visible(visible_layer, True)
                                    else:
                                        pdb.gimp_item_set_visible(visible_layer, False)
                                    multi_loop += 1
					
                        ####temp marker
                        # copy the contents of the visible layers into the selection
                        pdb.gimp_selection_grow(image, 1)
                        success = pdb.gimp_edit_copy_visible(image)
				
                        # turn on layer visibility
                        pdb.gimp_item_set_visible(layer_copy, True) 
					
                        # paste the selection into the floating layer
                        floating_sel = pdb.gimp_edit_paste(layer_copy,True)
                        # and anchor it
                        pdb.gimp_floating_sel_anchor(floating_sel)
                        # set the layer mode as per the dialog UI when the contents of the original character layer have not been cleared
                        if not clear_contents: pdb.gimp_layer_set_mode(layer_copy, mode)
                        # Re-position the copied layer
                        layer_copy.set_offsets(0, 0)
					
                        # use the original layer to copy the alpha selection before clearing the original character
                        if clear_contents: 
					
                            pdb.gimp_image_select_item(image, 2, layer_copy)
                            if grow>0:
                                pdb.gimp_selection_grow(image, grow)
                            else:
                                proc_trim_away(image, layer_copy, layer_names_list[visi_loop], 1)
					
					
                        # Make the original character layer visible before we merge down
					
					
					
					
                        pdb.gimp_item_set_visible( layer_names_list[visi_loop], True)
                        # option to clear the original character
                        if clear_contents: 
                            pdb.gimp_selection_none(image)
                            pdb.gimp_drawable_edit_clear(layer_names_list[visi_loop])
						
						
						
                        if not clear_contents: pdb.gimp_layer_set_opacity(layer_copy, opacity)
                        # Merge down the copied layer to the character layer
                        pdb.gimp_image_merge_down(image, layer_copy, 0)
                        # Re-set the id of the layer
                        layer_names_list[visi_loop] = pdb.gimp_image_get_active_layer(image)
                        # Remove annoying halo
                        #if not clear_contents: proc_trim_away(image, layer_names_list[visi_loop], layer_names_list[visi_loop], 1)
                        # Turn off layer visibility
                        pdb.gimp_item_set_visible(layer_names_list[visi_loop], False) 
				
				
                
                    visi_loop += 1
				
                    if has_a_space:
                        # re-position the space layer (just in case)
                        layer_names_list[space_loop].set_offsets(0, 0)
		
                    # Deselect
                    pdb.gimp_selection_none(image)
				
                ######
				
            visi_loop = 0
            # Re-set the list of layer ids
            layer_names_list = image.layers
        
            while visi_loop < num_layers and other_layers > 0:
                visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
                visi_name = visi_name.decode('utf-8')
                if visi_loop <> first_letter_layer:
                    # turn off visibility of letter layers
                    pdb.gimp_item_set_visible(layer_names_list[visi_loop],False)  
                visi_loop += 1
				
               
            # turn on visibility of character at the top of the layer stack and make it the active layer
            pdb.gimp_item_set_visible(layer_names_list[first_letter_layer],True)
            pdb.gimp_image_set_active_layer(image, layer_names_list[first_letter_layer])
				
        else:
            message_text = "There appears to be no suitable layer to create new letters from."
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message(message_text)
            pdb.gimp_message_set_handler(old_handler)            
			
		

    if xcffilename == "": do_merge_files = False
	
	
    if do_merge_files:
        cf_layers, visi_loop = 0, 0
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_names_list = image.layers
        while visi_loop < num_layers:# loop through layers and store any non-character layers for creating new faces
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if visi_name == "space.png" or visi_name == " .png": has_a_space = True
            if fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name): # check for custom font layer now allows unicode symbols
                cf_layers += 1
                if cf_layers>0: 
                    visi_loop = num_layers                    		
            visi_loop += 1
        if cf_layers == 0: 
            do_merge_files = False
            message_text = "There appears to be no suitable character layers."
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message(message_text)
            pdb.gimp_message_set_handler(old_handler)     
	
    if do_merge_files:
        mode = fn_get_mode(old_mode, new_mode)
        font_name_pointer, file_dummy = -1, 0
        fontimg = pdb.gimp_xcf_load(file_dummy,xcffilename,xcffilename)
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_list = image.layers
        loop, first_letter_layer = 0, -1
		
		
	
        if do_preview:
            my_first_letter_layer = pdb.gimp_image_get_active_layer(image)
            preview_other, preview_case, preview_shape, preview_vowel = 0, 0, 0, 0
		
            preview_im_width, preview_im_height = image.width, image.height
            preview_image = pdb.gimp_image_new(preview_im_width, preview_im_height, 0)
            temp_preview_layer = pdb.gimp_layer_new(preview_image, preview_im_width, preview_im_height, 1, "temp preview layer", 100, 28)
            pdb.gimp_image_insert_layer(preview_image, temp_preview_layer, None, -1)
            pdb.gimp_image_remove_layer(preview_image, temp_preview_layer)
            preview_display = pdb.gimp_display_new(preview_image)
            # Set start undo point for preview image
            pdb.gimp_image_undo_group_start(preview_image)	
            preview_layer_pos = -1
			
			
            visi_loop = 0
            real_name_layer_list = ["dummy"] # list to hold actual character names 
            while visi_loop < num_layers: # turn off visibility of letter layers
                real_visi_name = pdb.gimp_item_get_name(layer_list[visi_loop])
                visi_name = real_visi_name.decode('utf-8')
                if visi_name == "space.png" or visi_name == " .png": 
                    has_a_space = True
                    space_loop = visi_loop
                if len(visi_name) == 1 or len(visi_name) == 5 and ".png" in visi_name:
                    pdb.gimp_item_set_visible(layer_list[visi_loop],False) 
                    if real_visi_name<>"space.png" and real_visi_name<>" .png": 
                        real_name_layer_list.append(real_visi_name)
                        if preview_layer_pos == -1: preview_layer_pos = visi_loop
					
                visi_loop += 1
				
				
            preview_layer = layer_list[preview_layer_pos]
            preview_layer_name = pdb.gimp_item_get_name(preview_layer)
            temp_preview_layer = pdb.gimp_layer_new(preview_image, preview_layer.width, preview_layer.height, 1,  str(preview_layer_name), 100, 28)
            pdb.gimp_image_insert_layer(preview_image, temp_preview_layer, None, -1)
            success = pdb.gimp_edit_copy(preview_layer)
            floating_sel = pdb.gimp_edit_paste(temp_preview_layer, True)
            pdb.gimp_floating_sel_anchor(floating_sel)
            pdb.gimp_item_set_visible(temp_preview_layer,False) 
			
			
            custom_layer = pdb.gimp_image_get_layer_by_name(fontimg, preview_layer_name)
			
            if custom_layer >= 0: #layer found
                pdb.gimp_item_set_visible(temp_preview_layer,True)
                pdb.gimp_image_set_active_layer(preview_image, temp_preview_layer)
                success = pdb.gimp_edit_copy(custom_layer)
				
				
                if fixed_position: # copies directly onto existing layer with same character name
                    floating_sel = pdb.gimp_edit_paste(temp_preview_layer, False)
                    pdb.gimp_layer_set_opacity(floating_sel, opacity)
                else: # creates a new layer to copy to 
                    temp_layer = gimp.Layer(preview_image, "Temp Layer", custom_layer.width, custom_layer.height, RGBA_IMAGE, 100, 0) # layer 
                    preview_image.add_layer(temp_layer, -1)
                    floating_sel = pdb.gimp_edit_paste(temp_layer, False)

                if not clear_contents: pdb.gimp_layer_set_mode(floating_sel, mode)	
				
                pdb.gimp_floating_sel_anchor(floating_sel)
				
                if not fixed_position:
                    pdb.gimp_layer_set_opacity(temp_layer, opacity)
                    real_name_layer_list[loop] = pdb.gimp_image_merge_down(preview_image, temp_layer, 2)
                pdb.gimp_item_set_visible(preview_layer,False)
                #pdb.gimp_item_set_visible(preview_layer,True)
		
            # Set end undo point for preview image
            pdb.gimp_image_undo_group_end(preview_image)
            if my_first_letter_layer >-1:pdb.gimp_item_set_visible(my_first_letter_layer,True)
        else:
		
            while loop < num_layers:
                layer_name = pdb.gimp_item_get_name(layer_list[loop])
                custom_layer = pdb.gimp_image_get_layer_by_name(fontimg, layer_name)
                if (len(layer_name) == 5  and ".png" in layer_name) or (len(layer_name) == 6 and ".png" in layer_name):
                    if first_letter_layer <0: first_letter_layer = loop
                if custom_layer >= 0: #layer found
                    pdb.gimp_item_set_visible(layer_list[loop],True)
                    pdb.gimp_image_set_active_layer(image, layer_list[loop])
                    success = pdb.gimp_edit_copy(custom_layer)
                    if fixed_position:
                        floating_sel = pdb.gimp_edit_paste(layer_list[loop], False)
                        pdb.gimp_layer_set_opacity(floating_sel, opacity)
					
                    else:
                        temp_layer = gimp.Layer(image, "Temp Layer", custom_layer.width, custom_layer.height, RGBA_IMAGE, 100, 0) # layer 
                        image.add_layer(temp_layer, -1)
                        floating_sel = pdb.gimp_edit_paste(temp_layer, False)
                    #success = pdb.gimp_edit_copy(custom_layer)
                    # set the layer mode as per the dialog UI when the contents of the original character layer have not been cleared
                    if not clear_contents: pdb.gimp_layer_set_mode(floating_sel, mode)
                    pdb.gimp_floating_sel_anchor(floating_sel)
                    
                    if not fixed_position: 
                        pdb.gimp_layer_set_opacity(temp_layer, opacity)
                        layer_list[loop] = pdb.gimp_image_merge_down(image, temp_layer, 2)
                    pdb.gimp_item_set_visible(layer_list[loop],False)
                loop += 1
        
            if first_letter_layer >-1:pdb.gimp_item_set_visible(layer_list[first_letter_layer],True)
        
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
		
######################################################################################################################################

def fn_copy_list(new_list): 
    copy_list = new_list[:] 
    return copy_list 
	
######################################################################################################################################

def fn_get_mode(old_mode, new_mode):
        # Use Legacy Modes if new Default Modes are set to Normal Mode
        temp_mode = 0
        if new_mode <> 0:
            if new_mode == 1:
                temp_mode = 1
            elif new_mode == 2:
                temp_mode = 57
            elif new_mode == 3:
                temp_mode = 59
            elif new_mode == 4:
                temp_mode = 60
            elif new_mode == 5:
                temp_mode = 36
            elif new_mode == 6:
                temp_mode = 55
            elif new_mode == 7:
                temp_mode = 31
            elif new_mode == 8:
                temp_mode = 42
            elif new_mode == 9:
                temp_mode = 33
            elif new_mode == 10:
                temp_mode = 35
            elif new_mode == 11:
                temp_mode = 54
            elif new_mode == 12:
                temp_mode = 30
            elif new_mode == 13:
                temp_mode = 43
            elif new_mode == 14:
                temp_mode = 53
            elif new_mode == 15:
                temp_mode = 23
            elif new_mode == 16:
                temp_mode = 45
            elif new_mode == 17:
                temp_mode = 44
            elif new_mode == 18:
                temp_mode = 48
            elif new_mode == 19:
                temp_mode = 49
            elif new_mode == 20:
                temp_mode = 50
            elif new_mode == 21:
                temp_mode = 51
            elif new_mode == 22:
                temp_mode = 32
            elif new_mode == 23:
                temp_mode = 52
            elif new_mode == 24:
                temp_mode = 34
            elif new_mode == 25:
                temp_mode = 46
            elif new_mode == 26:
                temp_mode = 47
            elif new_mode == 27:
                temp_mode = 41
            elif new_mode == 28:
                temp_mode = 37
            elif new_mode == 29:
                temp_mode = 38
            elif new_mode == 30:
                temp_mode = 39
            elif new_mode == 31:
                temp_mode = 40
            elif new_mode == 32:
                temp_mode = 24
            elif new_mode == 33:
                temp_mode = 25
            elif new_mode == 34:
                temp_mode = 26
            elif new_mode == 35:
                temp_mode = 27
            elif new_mode == 36:
                temp_mode = 56
            mode = temp_mode
        else:
            if old_mode == 1:
                temp_mode = 1
            elif old_mode == 2:
                temp_mode = 10
            elif old_mode == 3:
                temp_mode = 4
            elif old_mode == 4:
                temp_mode = 16
            elif old_mode == 5:
                temp_mode = 7
            elif old_mode == 6:
                temp_mode = 9
            elif old_mode == 7:
                temp_mode = 3
            elif old_mode == 8:
                temp_mode = 17
            elif old_mode == 9:
                temp_mode = 5
            elif old_mode == 10:
                temp_mode = 19
            elif old_mode == 11:
                temp_mode = 18
            elif old_mode == 12:
                temp_mode = 6
            elif old_mode == 13:
                temp_mode = 8
            elif old_mode == 14:
                temp_mode = 20
            elif old_mode == 15:
                temp_mode = 21
            elif old_mode == 16:
                temp_mode = 22
            elif old_mode == 17:
                temp_mode = 11
            elif old_mode == 18:
                temp_mode = 12
            elif old_mode == 19:
                temp_mode = 13
            elif old_mode == 20:
                temp_mode = 14
            else:
                temp_mode = 0
            mode = temp_mode
			

        return mode
	
def proc_trim_away(img, l_copy, layer, amount):
    # only use when clearing the contents of the original character
    # set our feather settings
    our_feather = True
    our_feather_radius = 1
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # Comment out the shrink for use when clearing the contents of the original character
    # shrink the selection
    #pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(l_copy)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
######################################################################################################################################

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    #pdb.gimp_message("test = " + str(test) )
    return test
	
######################################################################################################################################


register(
    "python_fu_Custom_Font_Face_Maker_v0-10",
    "Provides options for making custom font faces",
    "Provides options for making custom font faces",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2022",
    "<Image>/Filters/Custom Font Tools/Custom Font Face Maker v0.10...",
    "*",
    [
        (PF_TOGGLE, "random_bg_option", "Create new characters \nfrom visible layers:", 1),
        (PF_TOGGLE, "fixed_bg_option", "Fixed position \nfrom visible layers:", 0),
        (PF_SPINNER, "pf_grow", "Grow characters within layers: ", 0.0,(0.0,10.0,0.2)),
        (PF_TOGGLE, "multi_bg_option", "Use all available\n decorative layers:", 0),
        (PF_TOGGLE, "multi_rnd_option", "Randomise use of\n decorative layers:", 0),
        (PF_OPTION, "group_option", "Character groups", 0, ("None used", "By case (upper, lower, number, punctuation)", "By shape (straight, curved, mixed)", "By vowel (vowel, consonant)")),
        (PF_TOGGLE, "clear_face_option", "Clear contents of \noriginal character:", 0),
        (PF_OPTION, "mode_option", "Legacy Mode \nto use with font face:", 0, ("Normal", "Dissolve", "Lighten", "Screen", "Dodge", "Addition", "Darken", "Multiply", "Burn", "Overlay", "Softlight", "Hardlight", "Difference", "Subtract", "Grain Extract", "Grain Merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSV)")),
        (PF_OPTION, "mode_option2", "New Mode \nto use with font face:", 0, ("Normal", "Dissolve","Color erase", "Merge", "Split", "Lighten only", "Luma/Luminance lighten only", "Screen", "Dodge", "Addition", "Darken only", "Luma/Luminance darken only", "Multiply", "Burn", "Linear burn", "Overlay", "Soft light", "Hard light", "Vivid light", "Pin light", "Linear light", "Hard mix", "Difference", "Exclusion", "Subtract", "Grain extract", "Grain merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSL)", "Hue (LCH)", "Chroma (LCH)", "Color (LCH)", "Lightness (LCH)", "Luminance")),
        (PF_SPINNER, "pf_opacity", "Opacity: ", 100,(5,100,1)),
        (PF_TOGGLE, "pf_usecustomfont", "Combine the current character \nwith one from the custom font below:", 0),
        (PF_FILENAME, "pf_names_xcffilename", ".xcf custom font file:", 0),
        (PF_TOGGLE, "do_preview", "Preview characters\n(to a new image): ", 0),
    ],
    [],
    do_stuff)

main()

	

